@@ -83,7 +83,7 @@ function nf_strip(self, level, target)
     ,   all   = "-s"
     }
     if self:is_plat("macosx", "iphoneos") then
-        maps.all = "-Wl,-x"
+        maps.all = {"-Wl,-x", "-Wl,-dead_strip"}
     end
     return maps[level]
 end
@@ -181,6 +181,8 @@ function nf_language(self, stdname)
             ansi        = "-ansi"
         ,   c89         = "-std=c89"
         ,   gnu89       = "-std=gnu89"
+        ,   c90         = "-std=c90"
+        ,   gnu90       = "-std=gnu90"
         ,   c99         = "-std=c99"
         ,   gnu99       = "-std=gnu99"
         ,   c11         = "-std=c11"
@@ -246,7 +248,7 @@ end
 
 -- make the define flag
 function nf_define(self, macro)
-    return "-D" .. macro
+    return {"-D" .. macro}
 end
 
 -- make the undefine flag
@@ -283,10 +285,15 @@ end
 function nf_rpathdir(self, dir)
     dir = path.translate(dir)
     if self:has_flags("-Wl,-rpath=" .. dir, "ldflags") then
-        return {"-Wl,-rpath=" .. (dir:gsub("@[%w_]+", function (name)
+        local flags = {"-Wl,-rpath=" .. (dir:gsub("@[%w_]+", function (name)
             local maps = {["@loader_path"] = "$ORIGIN", ["@executable_path"] = "$ORIGIN"}
             return maps[name]
         end))}
+        if self:is_plat("bsd") then
+            -- FreeBSD ld must have "-zorigin" with "-rpath".  Otherwise, $ORIGIN is not translated and it is literal.
+            table.insert(flags, 1, "-Wl,-zorigin")
+        end
+        return flags
     elseif self:has_flags("-Xlinker -rpath -Xlinker " .. dir, "ldflags") then
         return {"-Xlinker", "-rpath", "-Xlinker", (dir:gsub("%$ORIGIN", "@loader_path"))}
     end
@@ -316,7 +323,7 @@ end
 
 -- make the c precompiled header flag
 function nf_pcheader(self, pcheaderfile, target)
-    if self:kind() == "cc" then
+    if self:kind() == "cc" or self:kind() == "mm" then
         local pcoutputfile = target:pcoutputfile("c")
         if self:name() == "clang" then
             return {"-include", pcheaderfile, "-include-pch", pcoutputfile}
@@ -328,7 +335,7 @@ end
 
 -- make the c++ precompiled header flag
 function nf_pcxxheader(self, pcheaderfile, target)
-    if self:kind() == "cxx" then
+    if self:kind() == "cxx" or self:kind() == "mxx" then
         local pcoutputfile = target:pcoutputfile("cxx")
         if self:name() == "clang" then
             return {"-include", pcheaderfile, "-include-pch", pcoutputfile}
@@ -392,7 +399,7 @@ function link(self, objectfiles, targetkind, targetfile, flags)
     if option.get("verbose") then
         os.execv(program, argv, {envs = self:runenvs()})
     else
-        os.runv(program, argv, {envs = self:runenvs()})
+        os.vrunv(program, argv, {envs = self:runenvs()})
     end
 end
 
@@ -435,6 +442,8 @@ function _preprocess(program, argv, opt)
             is_gcc = true
         elseif tool:name():startswith("clang") then
             is_clang = true
+        elseif tool:name() == "circle" then
+            return
         end
     end
 
