@@ -20,31 +20,82 @@
 
 -- imports
 import("core.base.option")
+import("core.base.json")
+import("core.base.semver")
 import("core.tool.compiler")
 import("core.project.project")
 import("core.project.depend")
 import("core.project.config")
+import("lib.detect.find_tool")
 import("utils.progress")
 import("private.action.build.object", {alias = "objectbuilder"})
 import("common")
 import("stl_headers")
 
+-- get clang version
+function _get_clang_version(target)
+    local clang_version = _g.clang_version
+    if not clang_version then
+        local program, toolname = target:tool("cxx")
+        if program and (toolname == "clang" or toolname == "clangxx") then
+            local clang = find_tool("clang", {program = program, version = true})
+            if clang then
+                clang_version = clang.version
+            end
+        end
+        clang_version = clang_version or false
+        _g.clang_version = clang_version
+    end
+    return clang_version or nil
+end
+
+-- get clang-scan-deps
+function _get_clang_scan_deps(target)
+    local clang_scan_deps = _g.clang_scan_deps
+    if not clang_scan_deps then
+        local program, toolname = target:tool("cxx")
+        if program and (toolname == "clang" or toolname == "clangxx") then
+            local dir = path.directory(program)
+            local basename = path.basename(program)
+            local extension = path.extension(program)
+            program = (basename:gsub("clang", "clang-scan-deps")) .. extension
+            if os.isdir(dir) then
+                program = path.join(dir, program)
+            end
+            local result = find_tool("clang-scan-deps", {program = program, version = true})
+            if result then
+                clang_scan_deps = result.program
+            end
+        end
+        clang_scan_deps = clang_scan_deps or false
+        _g.clang_scan_deps = clang_scan_deps
+    end
+    return clang_scan_deps or nil
+end
+
 -- add a module or an header unit into the mapper
 --
 -- e.g
 -- -fmodule-file=build/.gens/Foo/rules/modules/cache/foo.pcm
 -- -fmodule-file=build/.gens/Foo/rules/modules/cache/iostream.pcm
 -- -fmodule-file=build/.gens/Foo/rules/modules/cache/bar.hpp.pcm
+-- on LLVM >= 16
+-- -fmodule-file=foo=build/.gens/Foo/rules/modules/cache/foo.pcm
+-- -fmodule-file=build/.gens/Foo/rules/modules/cache/iostream.pcm
+-- -fmodule-file=build/.gens/Foo/rules/modules/cache/bar.hpp.pcm
 --
-function _add_module_to_mapper(target, name, bmifile, deps)
+function _add_module_to_mapper(target, name, bmifile, opt)
+    opt = opt or {}
     local modulemap = _get_modulemap_from_mapper(target, name)
     if modulemap then
         return
     end
 
+    local clang_version = _get_clang_version(target)
+    local namedmodule = opt.namedmodule and semver.compare(clang_version, "16.0") >= 0
     local modulefileflag = get_modulefileflag(target)
-    local mapflag = modulefileflag .. bmifile
-    modulemap = {flag = mapflag, deps = deps}
+    local mapflag = namedmodule and format("%s%s=%s", modulefileflag, name, bmifile) or modulefileflag .. bmifile
+    modulemap = {flag = mapflag, deps = opt.deps}
     common.localcache():set2(_mapper_cachekey(target), "modulemap" .. name, modulemap)
 end
 
@@ -64,31 +115,51 @@ function _get_modulemap_from_mapper(target, name)
     return common.localcache():get2(_mapper_cachekey(target), "modulemap" .. name) or nil
 end
 
+-- use the given stdlib? e.g. libc++ or libstdc++
+function _use_stdlib(target, name)
+    local stdlib = target:data("cxx.modules.stdlib") or "libstdc++"
+    return stdlib == name
+end
+
+-- set stdlib flags, it will use libstdc++ if we do not set `-stdlib=`
+function _set_stdlib_flags(target)
+    if _use_stdlib(target, "libc++") then
+        target:add("cxxflags", "-stdlib=libc++")
+        target:add("ldflags", "-stdlib=libc++")
+        target:add("shflags", "-stdlib=libc++")
+    end
+end
+
 -- load module support for the current target
 function load(target)
-    local modulesflag = get_modulesflag(target)
-    local builtinmodulemapflag = get_builtinmodulemapflag(target)
-    local implicitmodulesflag = get_implicitmodulesflag(target)
+    local clangmodulesflag, modulestsflag, withoutflag = get_modulesflag(target)
 
     -- add module flags
-    target:add("cxxflags", modulesflag)
-    target:add("cxxflags", builtinmodulemapflag, {force = true})
-    target:add("cxxflags", implicitmodulesflag, {force = true})
+    if not withoutflag then
+        target:add("cxxflags", modulestsflag)
+    end
+
+    -- enable clang modules to emulate std modules
+    if target:policy("build.c++.clang.stdmodules") then
+       target:add("cxxflags", clangmodulesflag)
+    end
 
     -- fix default visibility for functions and variables [-fvisibility] differs in PCH file vs. current file
     -- module.pcm cannot be loaded due to a configuration mismatch with the current compilation.
     --
-    -- it will happen in binary target depend ont shared target with modules, and enable release mode at same time.
+    -- it will happen in binary target depend on library target with modules, and enable release mode at same time.
+    --
+    -- @see https://github.com/xmake-io/xmake/issues/3358#issuecomment-1432586767
     local dep_symbols
-    local has_shared_deps = false
+    local has_library_deps = false
     for _, dep in ipairs(target:orderdeps()) do
-        if dep:is_shared() then
+        if dep:is_shared() or dep:is_static() or dep:is_object() then
             dep_symbols = dep:get("symbols")
-            has_shared_deps = true
+            has_library_deps = true
             break
         end
     end
-    if has_shared_deps then
+    if has_library_deps then
         target:set("symbols", dep_symbols and dep_symbols or "none")
     end
 
@@ -97,10 +168,13 @@ function load(target)
     -- on ubuntu:
     -- sudo apt install libc++-dev libc++abi-15-dev
     --
-    target:data_set("cxx.modules.use_libc++", table.contains(target:get("cxxflags"), "-stdlib=libc++", "clang::-stdlib=libc++"))
-    if target:data("cxx.modules.use_libc++") then
-        target:add("syslinks", "c++")
+    local flags = table.join(target:get("cxxflags") or {}, get_config("cxxflags") or {})
+    if table.contains(flags, "-stdlib=libc++", "clang::-stdlib=libc++") then
+        target:data_set("cxx.modules.stdlib", "libc++")
+    elseif table.contains(flags, "-stdlib=libstdc++", "clang::-stdlib=libstdc++") then
+        target:data_set("cxx.modules.stdlib", "libstdc++")
     end
+    _set_stdlib_flags(target)
 end
 
 -- get includedirs for stl headers
@@ -114,10 +188,10 @@ function _get_toolchain_includedirs_for_stlheaders(target, includedirs, clang)
     local tmpfile = os.tmpfile() .. ".cc"
     io.writefile(tmpfile, "#include <vector>")
     local argv = {"-E", "-x", "c++", tmpfile}
-    if target:data("cxx.modules.use_libc++") then
+    if _use_stdlib(target, "libc++") then
         table.insert(argv, 1, "-stdlib=libc++")
     end
-    local result = try {function () return os.iorunv(clang, {"-E", "-x", "c++", tmpfile}) end}
+    local result = try {function () return os.iorunv(clang, argv) end}
     if result then
         for _, line in ipairs(result:split("\n", {plain = true})) do
             line = line:trim()
@@ -135,9 +209,9 @@ end
 
 -- do compile for batchcmds
 -- @note we need use batchcmds:compilev to translate paths in compflags for generator, e.g. -Ixx
-function _batchcmds_compile(batchcmds, target, flags)
+function _batchcmds_compile(batchcmds, target, sourcefile, flags)
     local compinst = target:compiler("cxx")
-    local compflags = compinst:compflags({target = target})
+    local compflags = compinst:compflags({sourcefile = sourcefile, target = target})
     batchcmds:compilev(table.join(compflags or {}, flags), {compiler = compinst, sourcekind = "cxx"})
 end
 
@@ -146,7 +220,7 @@ function _build_modulefile(target, sourcefile, opt)
     local objectfile = opt.objectfile
     local dependfile = opt.dependfile
     local compinst = compiler.load("cxx", {target = target})
-    local compflags = compinst:compflags({target = target})
+    local compflags = compinst:compflags({sourcefile = sourcefile, target = target})
     local dependinfo = option.get("rebuild") and {} or (depend.load(dependfile) or {})
 
     -- need build this object?
@@ -160,29 +234,40 @@ function _build_modulefile(target, sourcefile, opt)
     -- init flags
     local common_args = opt.common_args
     local requiresflags = opt.requiresflags
+    local moduleoutputflag = get_moduleoutputflag(target)
 
     -- trace
     progress.show(opt.progress, "${color.build.object}compiling.module.$(mode) %s", opt.provide and opt.provide.name or sourcefile)
 
     local bmifile
+    local compileflags = {}
     local bmiflags
     if opt.provide then
         bmifile = opt.provide.bmifile
-        bmiflags = table.join("-x", "c++-module", "--precompile", compflags, common_args, requiresflags or {})
+        if moduleoutputflag then
+            compileflags = table.join("-x", "c++-module", moduleoutputflag .. bmifile, requiresflags)
+        else
+            bmiflags = table.join("-x", "c++-module", "--precompile", compflags, common_args, requiresflags)
+        end
+    else
+        compileflags = {"-x", "c++"}
+    end
+
+    if bmiflags then
         vprint(compinst:compcmd(sourcefile, bmifile, {compflags = bmiflags, rawargs = true}))
     end
 
-    local objflags = table.join(compflags, common_args, requiresflags or {}, (bmifile == nil) and {"-x", "c++"} or {})
-    vprint(compinst:compcmd(bmifile or sourcefile, objectfile, {compflags = objflags, rawargs = true}))
+    compileflags = table.join2(compileflags, compflags, common_args, requiresflags or {})
+    vprint(compinst:compcmd(bmiflags and bmifile or sourcefile, objectfile, {compflags = compileflags, rawargs = true}))
 
     if not dryrun then
 
         -- do compile
         dependinfo.files = {}
-        if opt.provide then
+        if bmiflags then
             assert(compinst:compile(sourcefile, bmifile, {dependinfo = dependinfo, compflags = bmiflags}))
         end
-        assert(compinst:compile(bmifile or sourcefile, objectfile, {compflags = objflags}))
+        assert(compinst:compile(bmiflags and bmifile or sourcefile, objectfile, {compflags = compileflags}))
 
         -- update files and values to the dependent file
         dependinfo.values = depvalues
@@ -197,9 +282,9 @@ function toolchain_includedirs(target)
     if includedirs == nil then
         includedirs = {}
         local clang, toolname = target:tool("cxx")
-        assert(toolname == "clang")
+        assert(toolname:startswith("clang"))
         _get_toolchain_includedirs_for_stlheaders(target, includedirs, clang)
-        local _, result = try {function () return os.iorunv(clang, {"-E", "-Wp,-v", "-xc", os.nuldev()}) end}
+        local _, result = try {function () return os.iorunv(clang, {"-E", "-stdlib=libc++", "-Wp,-v", "-xc", os.nuldev()}) end}
         if result then
             for _, line in ipairs(result:split("\n", {plain = true})) do
                 line = line:trim()
@@ -218,7 +303,6 @@ end
 -- generate dependency files
 function generate_dependencies(target, sourcebatch, opt)
     local changed = false
-    local cachedir = common.modules_cachedir(target)
     for _, sourcefile in ipairs(sourcebatch.sourcefiles) do
         local dependfile = target:dependfile(sourcefile)
         depend.on_changed(function()
@@ -226,45 +310,74 @@ function generate_dependencies(target, sourcebatch, opt)
                 progress.show(opt.progress, "${color.build.object}generating.module.deps %s", sourcefile)
             end
 
-            local outputdir = path.translate(path.join(cachedir, path.directory(path.relative(sourcefile, projectdir))))
-            if not os.isdir(outputdir) then
-                os.mkdir(outputdir)
-            end
-
-            -- no support of p1689 atm
+            local outputdir = common.get_outputdir(target, sourcefile)
             local jsonfile = path.translate(path.join(outputdir, path.filename(sourcefile) .. ".json"))
-            common.fallback_generate_dependencies(target, jsonfile, sourcefile, function(file)
+            if has_clangscandepssupport(target) and not target:policy("build.c++.clang.fallbackscanner") then
+                local clangscandeps = _get_clang_scan_deps(target)
                 local compinst = target:compiler("cxx")
-                local defines = {}
-                for _, define in pairs(target:get("defines")) do
-                    table.insert(defines, "-D" .. define)
-                end
-                local includedirs = table.join({}, target:get("includedirs"))
-                for _, dep in ipairs(target:orderdeps()) do
-                    local includedir = dep:get("sysincludedirs") or dep:get("includedirs")
-                    if includedir then
-                        table.join2(includedirs, includedir)
-                    end
-                end
-                for _, pkg in pairs(target:pkgs()) do
-                    local includedir = pkg:get("sysincludedirs") or pkg:get("includedirs")
-                    if includedir then
-                        table.join2(includedirs, includedir)
-                    end
-                end
-                for i, includedir in pairs(includedirs) do
-                    includedirs[i] = "-I" .. includedir
-                end
-                local ifile = path.translate(path.join(outputdir, path.filename(file) .. ".i"))
-                os.vrunv(compinst:program(), table.join(includedirs, defines, {"-E", "-x", "c++", file, "-o", ifile}))
-                local content = io.readfile(ifile)
-                os.rm(ifile)
-                return content
-            end)
+                local compflags = compinst:compflags({sourcefile = sourcefile, target = target})
+                local flags = table.join("--format=p1689", "--",
+                                         compinst:program(), "-x", "c++", "-c", sourcefile, "-o", target:objectfile(sourcefile),
+                                         compflags)
+                vprint(table.concat(table.join(clangscandeps, flags), " "))
+                local outdata, errdata = os.iorunv(clangscandeps, flags)
+                assert(errdata, errdata)
+
+                io.writefile(jsonfile, outdata)
+            else
+                common.fallback_generate_dependencies(target, jsonfile, sourcefile, function(file)
+                    local compinst = target:compiler("cxx")
+                    local compflags = compinst:compflags({sourcefile = file, target = target})
+                    -- exclude -fmodule* and -std=c++/gnu++* flags because, when they are set clang try to find bmi of imported modules but they don't exists a this point of compilation
+                    table.remove_if(compflags, function(_, flag)
+                        return flag:startswith("-fmodule") or flag:startswith("-std=c++") or flag:startswith("-std=gnu++")
+                    end)
+                    local ifile = path.translate(path.join(outputdir, path.filename(file) .. ".i"))
+                    os.vrunv(compinst:program(), table.join(compflags, {"-E", "-x", "c++", file, "-o", ifile}))
+                    local content = io.readfile(ifile)
+                    os.rm(ifile)
+                    return content
+                end)
+            end
             changed = true
 
-            local dependinfo = io.readfile(jsonfile)
-            return {moduleinfo = dependinfo}
+            local rawdependinfo = io.readfile(jsonfile)
+            if rawdependinfo then
+                local dependinfo = json.decode(rawdependinfo)
+                if target:data("cxx.modules.stdlib") == nil then
+                    local has_std_modules = false
+                    for _, r in ipairs(dependinfo.rules) do
+                        for _, required in ipairs(r.requires) do
+                            -- it may be `std:utility`, ..
+                            -- @see https://github.com/xmake-io/xmake/issues/3373
+                            local logical_name = required["logical-name"]
+                            if logical_name and (logical_name == "std" or logical_name:startswith("std.") or logical_name:startswith("std:")) then
+                                has_std_modules = true
+                                break
+                            end
+                        end
+
+                        if has_std_modules then
+                            break
+                        end
+                    end
+                    if has_std_modules then
+
+                        -- we need clang >= 17.0 or use clang stdmodules if the current target contains std module
+                        local clang_version = _get_clang_version(target)
+                        assert((clang_version and semver.compare(clang_version, "17.0") >= 0) or target:policy("build.c++.clang.stdmodules"),
+                               [[On llvm <= 16 standard C++ modules are not supported ;
+                               they can be emulated through clang modules and supported only on libc++ ;
+                               please add -stdlib=libc++ cxx flag or disable strict mode]])
+
+                        -- we use libc++ by default if we do not explicitly specify -stdlib:libstdc++
+                        target:data_set("cxx.modules.stdlib", "libc++")
+                        _set_stdlib_flags(target)
+                    end
+                end
+            end
+
+            return {moduleinfo = rawdependinfo}
         end, {dependfile = dependfile, files = {sourcefile}})
     end
     return changed
@@ -283,7 +396,7 @@ function generate_stl_headerunits_for_batchjobs(target, batchjobs, headerunits,
     end, {rootjob = opt.rootjob})
 
     -- build headerunits
-    for i, headerunit in ipairs(headerunits) do
+    for _, headerunit in ipairs(headerunits) do
         local bmifile = path.join(stlcachedir, headerunit.name .. get_bmi_extension())
         if not os.isfile(bmifile) then
             batchjobs:addjob(headerunit.name, function (index, total)
@@ -292,13 +405,13 @@ function generate_stl_headerunits_for_batchjobs(target, batchjobs, headerunits,
                     if not common.memcache():get2(headerunit.name, "building") then
                         common.memcache():set2(headerunit.name, "building", true)
                         progress.show((index * 100) / total, "${color.build.object}compiling.headerunit.$(mode) %s", headerunit.name)
-                        local args = {modulecachepathflag .. stlcachedir, "-c", "-o", bmifile, "-x", "c++-system-header", headerunit.name}
+                        local args = {modulecachepathflag .. stlcachedir, "-c", "-Wno-everything", "-o", bmifile, "-x", "c++-system-header", headerunit.name}
                         os.vrunv(compinst:program(), table.join(compinst:compflags({target = target}), args))
                     end
 
                 end, {dependfile = target:dependfile(bmifile), files = {headerunit.path}})
                 -- libc++ have a builtin module mapper
-                if not target:data_set("cxx.modules.use_libc++") then
+                if not _use_stdlib(target, "libc++") then
                     _add_module_to_mapper(target, headerunit.name, bmifile)
                 end
             end, {rootjob = flushjob})
@@ -313,9 +426,8 @@ function generate_stl_headerunits_for_batchcmds(target, batchcmds, headerunits,
     assert(has_headerunitsupport(target), "compiler(clang): does not support c++ header units!")
 
     -- build headerunits
-    local projectdir = os.projectdir()
     local depmtime = 0
-    for i, headerunit in ipairs(headerunits) do
+    for _, headerunit in ipairs(headerunits) do
         local bmifile = path.join(stlcachedir, headerunit.name .. get_bmi_extension())
         -- don't build same header unit at the same time
         if not common.memcache():get2(headerunit.name, "building") then
@@ -327,7 +439,7 @@ function generate_stl_headerunits_for_batchcmds(target, batchcmds, headerunits,
             _batchcmds_compile(batchcmds, target, flags)
         end
         -- libc++ have a builtin module mapper
-        if not target:data_set("cxx.modules.use_libc++") then
+        if not _use_stdlib(target, "libc++") then
             _add_module_to_mapper(target, headerunit.name, bmifile)
         end
         depmtime = math.max(depmtime, os.mtime(bmifile))
@@ -342,7 +454,7 @@ function generate_user_headerunits_for_batchjobs(target, batchjobs, headerunits,
     assert(has_headerunitsupport(target), "compiler(clang): does not support c++ header units!")
 
     -- get cachedirs
-    local cachedir = common.modules_cachedir(target)
+    local cachedir = common.modules_cachedir(target, {mkdir = true})
     local modulecachepathflag = get_modulecachepathflag(target)
 
     -- flush job
@@ -351,19 +463,13 @@ function generate_user_headerunits_for_batchjobs(target, batchjobs, headerunits,
     end, {rootjob = opt.rootjob})
 
     -- build headerunits
-    local projectdir = os.projectdir()
     for _, headerunit in ipairs(headerunits) do
         local file = path.relative(headerunit.path, target:scriptdir())
         local objectfile = target:objectfile(file)
 
-        local outputdir
-        if headerunit.type == ":quote" then
-            outputdir = path.join(cachedir, path.directory(path.relative(headerunit.path, projectdir)))
-        else
-            outputdir = path.join(cachedir, path.directory(headerunit.path))
-        end
+        local outputdir = common.get_outputdir(target, headerunit.path)
         local bmifilename = path.basename(objectfile) .. get_bmi_extension()
-        local bmifile = (outputdir and path.join(outputdir, bmifilename) or bmifilename)
+        local bmifile = path.join(outputdir, bmifilename)
         batchjobs:addjob(headerunit.name, function (index, total)
             depend.on_changed(function()
                 progress.show((index * 100) / total, "${color.build.object}compiling.headerunit.$(mode) %s", headerunit.name)
@@ -395,22 +501,16 @@ function generate_user_headerunits_for_batchcmds(target, batchcmds, headerunits,
     assert(has_headerunitsupport(target), "compiler(clang): does not support c++ header units!")
 
     -- get cachedirs
-    local cachedir = common.modules_cachedir(target)
+    local cachedir = common.modules_cachedir(target, {mkdir = true})
     local modulecachepathflag = get_modulecachepathflag(target)
 
     -- build headerunits
-    local projectdir = os.projectdir()
     local depmtime = 0
     for _, headerunit in ipairs(headerunits) do
         local file = path.relative(headerunit.path, target:scriptdir())
         local objectfile = target:objectfile(file)
 
-        local outputdir
-        if headerunit.type == ":quote" then
-            outputdir = path.join(cachedir, path.directory(path.relative(headerunit.path, projectdir)))
-        else
-            outputdir = path.join(cachedir, path.directory(headerunit.path))
-        end
+        local outputdir = common.get_outputdir(target, headerunit.path)
         batchcmds:mkdir(outputdir)
 
         local bmifilename = path.basename(objectfile) .. get_bmi_extension()
@@ -471,6 +571,24 @@ function build_modules_for_batchjobs(target, batchjobs, objectfiles, modules, op
                 end
             end
             local moduleinfo = table.copy(provide) or {}
+
+            if provide then
+                local fileconfig = target:fileconfig(cppfile)
+                if fileconfig and fileconfig.install then
+                    batchjobs:addjob(name .. "_metafile", function(index, total)
+                        local outputdir = common.get_outputdir(target, cppfile)
+                        local metafilepath = path.join(outputdir, path.filename(cppfile) .. ".meta-info")
+                        depend.on_changed(function()
+                            progress.show(opt.progress, "${color.build.object}generating.module.metadata %s", name)
+                            local metadata = common.generate_meta_module_info(target, name, cppfile, module.requires)
+                            json.savefile(metafilepath, metadata)
+
+                        end, {dependfile = target:dependfile(metafilepath), files = {cppfile}})
+
+                    end, {rootjob = flushjob})
+                end
+            end
+
             table.join2(moduleinfo, {
                 name = name or cppfile,
                 deps = table.keys(module.requires or {}),
@@ -484,17 +602,22 @@ function build_modules_for_batchjobs(target, batchjobs, objectfiles, modules, op
 
                     if provide or common.has_module_extension(cppfile) then
                         local bmifile = provide and provide.bmi
-                        _build_modulefile(target, provide and provide.sourcefile or cppfile, {
-                            objectfile = objectfile,
-                            dependfile = target:dependfile(bmifile or objectfile),
-                            provide = provide and {bmifile = bmifile, name = name},
-                            common_args = common_args,
-                            requiresflags = requiresflags,
-                            progress = (index * 100) / total})
+                        if not common.memcache():get2(name or cppfile, "compiling") then
+                            if name and module.external then
+                                common.memcache():set2(name or cppfile, "compiling", true)
+                            end
+                            _build_modulefile(target, provide and provide.sourcefile or cppfile, {
+                                objectfile = objectfile,
+                                dependfile = target:dependfile(bmifile or objectfile),
+                                provide = provide and {bmifile = bmifile, name = name},
+                                common_args = common_args,
+                                requiresflags = requiresflags,
+                                progress = (index * 100) / total})
+                            end
                         target:add("objectfiles", objectfile)
 
                         if provide then
-                            _add_module_to_mapper(target, name, bmifile, requiresflags)
+                            _add_module_to_mapper(target, name, bmifile, {deps = requiresflags, namedmodule = true})
                         end
                     elseif requiresflags then
                         local cxxflags = {}
@@ -556,9 +679,9 @@ function build_modules_for_batchcmds(target, batchcmds, objectfiles, modules, op
                 if provide then
                     _batchcmds_compile(batchcmds, target, table.join(flags,
                         {"-x", "c++-module", "--precompile", "-c", path(cppfile), "-o", path(provide.bmi)}))
-                    _add_module_to_mapper(target, name, provide.bmi)
+                    _add_module_to_mapper(target, name, provide.bmi, {namedmodule = true})
                 end
-                _batchcmds_compile(batchcmds, target, table.join(flags,
+                _batchcmds_compile(batchcmds, target, file, table.join(flags,
                     not provide and {"-x", "c++"} or {}, {"-c", file, "-o", path(objectfile)}))
                 target:add("objectfiles", objectfile)
             elseif requiresflags then
@@ -593,21 +716,25 @@ function get_bmi_extension()
 end
 
 function get_modulesflag(target)
-    local modulesflag = _g.modulesflag
-    if modulesflag == nil then
+    local clangmodulesflag = _g.clangmodulesflag
+    local modulestsflag = _g.modulestsflag
+    local withoutflag = _g.withoutflag
+    if clangmodulesflag == nil and modulestsflag == nil then
         local compinst = target:compiler("cxx")
         if compinst:has_flags("-fmodules", "cxxflags", {flagskey = "clang_modules"}) then
-            modulesflag = "-fmodules"
+            clangmodulesflag = "-fmodules"
         end
-        if not modulesflag then
-            if compinst:has_flags("-fmodules-ts", "cxxflags", {flagskey = "clang_modules_ts"}) then
-                modulesflag = "-fmodules-ts"
-            end
+        if compinst:has_flags("-fmodules-ts", "cxxflags", {flagskey = "clang_modules_ts"}) then
+            modulestsflag = "-fmodules-ts"
         end
-        assert(modulesflag, "compiler(clang): does not support c++ module!")
-        _g.modulesflag = modulesflag or false
+        local clang_version = _get_clang_version(target)
+        withoutflag = semver.compare(clang_version, "16.0") >= 0
+        assert(withoutflag or modulestsflag, "compiler(clang): does not support c++ module!")
+        _g.clangmodulesflag = clangmodulesflag or false
+        _g.modulestsflag = modulestsflag or false
+        _g.withoutflag = withoutflag or false
     end
-    return modulesflag or nil
+    return clangmodulesflag or nil, modulestsflag or nil, withoutflag or nil
 end
 
 function get_builtinmodulemapflag(target)
@@ -640,6 +767,19 @@ function get_implicitmodulesflag(target)
     return implicitmodulesflag or nil
 end
 
+function get_implicitmodulemapsflag(target)
+    local implicitmodulemapsflag = _g.implicitmodulemapsflag
+    if implicitmodulemapsflag == nil then
+        local compinst = target:compiler("cxx")
+        if compinst:has_flags("-fimplicit-module-maps", "cxxflags", {flagskey = "clang_implicit_module_map"}) then
+            implicitmodulemapsflag = "-fimplicit-module-maps"
+        end
+        assert(implicitmodulemapsflag, "compiler(clang): does not support c++ module!")
+        _g.implicitmodulemapsflag = implicitmodulemapsflag or false
+    end
+    return implicitmodulemapsflag or nil
+end
+
 function get_noimplicitmodulemapsflag(target)
     local noimplicitmodulemapsflag = _g.noimplicitmodulemapsflag
     if noimplicitmodulemapsflag == nil then
@@ -696,8 +836,16 @@ function has_headerunitsupport(target)
     local support_headerunits = _g.support_headerunits
     if support_headerunits == nil then
         local compinst = target:compiler("cxx")
-        if compinst:has_flags(get_modulesflag(target) .. " -std=c++20 -x c++-user-header", "cxxflags", {flagskey = "clang_user_header_unit_support", tryrun = true}) and
-           compinst:has_flags(get_modulesflag(target) .. " -std=c++20 -x c++-system-header", "cxxflags", {flagskey = "clang_system_header_unit_support", tryrun = true}) then
+        local _, modulestsflag, withoutflag = get_modulesflag(target)
+        modulestsflag = withoutflag and "" or modulestsflag
+        if compinst:has_flags(modulestsflag .. " -std=c++20 -x c++-user-header", "cxxflags", {
+                snippet = "inline int foo() { return 0; }",
+                flagskey = "clang_user_header_unit_support",
+                tryrun = true}) and
+           compinst:has_flags(modulestsflag .. " -std=c++20 -x c++-system-header", "cxxflags", {
+                snippet = "inline int foo() { return 0; }",
+                flagskey = "clang_system_header_unit_support",
+                tryrun = true}) then
             support_headerunits = true
         end
         _g.support_headerunits = support_headerunits or false
@@ -705,6 +853,33 @@ function has_headerunitsupport(target)
     return support_headerunits or nil
 end
 
+function has_clangscandepssupport(target)
+    local support_clangscandeps = _g.support_clangscandeps
+    if support_clangscandeps == nil then
+        local clangscandeps = _get_clang_scan_deps(target)
+        local clang_version = _get_clang_version(target)
+        if clangscandeps and clang_version and semver.compare(clang_version, "16.0") >= 0 then
+            support_clangscandeps = true
+        end
+        _g.support_clangscandeps = support_clangscandeps or false
+    end
+    return support_clangscandeps or nil
+end
+
+function get_moduleoutputflag(target)
+    local moduleoutputflag = _g.moduleoutputflag
+    if moduleoutputflag == nil then
+        local compinst = target:compiler("cxx")
+        local clang_version = _get_clang_version(target)
+        if compinst:has_flags("-fmodule-output=", "cxxflags", {flagskey = "clang_module_output", tryrun = true}) and
+            semver.compare(clang_version, "16.0") >= 0 then
+            moduleoutputflag = "-fmodule-output="
+        end
+        _g.moduleoutputflag = moduleoutputflag or false
+    end
+    return moduleoutputflag or nil
+end
+
 function get_requiresflags(target, requires)
     local flags = {}
     -- add deps required module flags
@@ -721,7 +896,7 @@ function get_requiresflags(target, requires)
                 already_mapped_modules[name] = true
                 table.insert(flags, modulemap_.flag)
                 if modulemap_.deps then
-                    table.shallow_join2(flags, modulemap_.deps)
+                    table.join2(flags, modulemap_.deps)
                 end
                 goto continue
             end
@@ -733,7 +908,7 @@ function get_requiresflags(target, requires)
             already_mapped_modules[name] = true
             table.insert(flags, modulemap.flag)
             if modulemap.deps then
-                table.shallow_join2(flags, modulemap.deps)
+                table.join2(flags, modulemap.deps)
             end
             goto continue
         end
